#!/usr/bin/perl
###########################################################################
# LSB Function Signature Checker v1.5, tool for checking ABI compliance
# of two Linux distributions using ABI Compliance Checker.
# Copyright (C) The Linux Foundation
# Copyright (C) Institute for System Programming, RAS
# Author: Andrey Ponomarenko
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
use Getopt::Long;
use File::Path qw(mkpath rmtree);
use FindBin;
sub BEGIN {
 unshift @INC, $FindBin::Bin;
}
use SimpleJournal;
Getopt::Long::Configure ("posix_default", "no_ignore_case");

my ($Help, $Run_Tests, $Create_TestSuite, $Check_Devel_Files, $Distribution_Name, $Source, $TargetLibrary);

my $CmdName = get_FileName($0);
GetOptions("help|h!" => \$Help,
  "d|distribution=s" => \$Distribution_Name,
  "src|source=s" => \$Source,
  "create!" => \$Create_TestSuite,
  "run!" => \$Run_Tests,
  "check!" => \$Check_Devel_Files,
  "l|library=s" => \$TargetLibrary);

sub HELP_MESSAGE()
{
    print STDERR <<"EOM"

NAME:
  $CmdName: check ABI compliance of two Linux distributions using ABI-compliance-checker.

USAGE:
  $CmdName [options]

EXAMPLES:
  $CmdName -create -src moblin-sigchk-data -d Moblin-2.1
  $CmdName -run -src moblin-sigchk-data -d Ubuntu-Moblin-Remix-9.10
  $CmdName -check -src moblin-sigchk-data -d Ubuntu-Moblin-Remix-9.10

OPTIONS:
  -h|-help
      Print help message.

  -d|-distribution <name>
      The name of distribution.

  -src|-source <path>
      Path to the test data.

  -create
      Create test suite (1st distribution).

  -run
      Run tests (2nd distribution).

  -l|-library <library>
      This option allow to recreate and run test case for the specified library.

  -check
      Check installed header files and shared objetcs.

EOM
      ;
}

my %TestResult;
my %Descriptor;
my (%DevelFiles, %DevelFileName_Paths);
my $TESTSUITE_DISTRIBUTION;
my (%LibPath_SoName, %LibName_SoName);
my %Cache;

sub recreate_test_case($)
{
    my $Library = $_[0];
    my $D_Path = $Source."/descriptors/$Library\.desc";
    my $Int_List_Path = $Source."/int_list/$Library\.ints";
    print "dumping $Library ABI\n";
    print "running abi-compliance-checker.pl -l \'$Library\' -dump_info \'$D_Path\' -int_list \'$Int_List_Path\'\n";
    system("$FindBin::Bin/abi-compliance-checker.pl -l \'$Library\' -dump_info \'$D_Path\' -int_list \'$Int_List_Path\'");
    print "\n";
    mkpath($Source."/test_data/".$Library);
    system("mv -f abi_dumps/$Library/* ".esc($Source)."/test_data/$Library/");
    rmtree("abi_dumps/");
}

sub esc($)
{
    my $Str = $_[0];
    $Str =~ s/([()\[\]{}$ &'"`;])/\\$1/g;
    return $Str;
}

sub cmd_find($$$)
{
    my ($Path, $Type, $Name) = @_;
    return () if(not $Path or not -e $Path);
    my $Cmd = "find ".esc($Path);
    if($Type)
    {
        $Cmd .= " -type $Type";
    }
    if($Name)
    {
        $Cmd .= " -name \"$Name\"";
    }
    return split("\n", `$Cmd`);
}

sub cmd_file($)
{
    my $Path = $_[0];
    return "" if(not $Path or not -e $Path);
    my $Cmd = "file ".esc($Path);
    my $Cmd_Out = `$Cmd`;
    return $Cmd_Out;
}

sub create_test_suite()
{#Dump ABI information using Moblin library descriptors
    rmtree("abi_dumps");
    my $LIB_SONAME = "";
    foreach my $D_Path (sort (cmd_find($Source."/descriptors","f","")))
    {
        $D_Path=~/\/([^\/]*)\.desc\Z/;
        my $Library = $1;
        my $Int_List_Path = $Source."/int_list/".$Library.".ints";
        print "dumping $Library ABI\n";
        print "running abi-compliance-checker.pl -l \'$Library\' -dump_info \'$D_Path\' -int_list \'$Int_List_Path\'\n";
        system("$FindBin::Bin/abi-compliance-checker.pl -l \'$Library\' -dump_info \'$D_Path\' -int_list \'$Int_List_Path\'");
        print "\n";
        readDescriptor($Library, $D_Path);
        foreach my $Lib_Path (split("\n", $Descriptor{$Library}{"Libs"}))
        {
            $Lib_Path=~s/\A[ ]+//g;
            $Lib_Path=~s/[ ]+\Z//g;
            $LIB_SONAME .= $Lib_Path.";".get_soname($Lib_Path)."\n";
        }
    }
    writeFile($Source."/sonames", $LIB_SONAME);
    rmtree($Source."/test_data/");
    system("mv", "abi_dumps", $Source."/test_data");
    writeFile($Source."/distribution.name", $Distribution_Name);
    print "test data for $Distribution_Name has been created in the directory '$Source/test_data/'\n";
}

sub check_devel_files()
{#Check system for available devel files
    mkpath("results/");
    my @Descriptors = cmd_find($Source."/descriptors","f","");
    my $NOT_INSTALLED = "";
    print "checking for devel files\n";
    my ($Success, $Fail) = (0, 0);
    foreach my $D_Path (sort {lc($a) cmp lc($b)} @Descriptors)
    {
        $D_Path =~ /\/([^\/]*)\.desc\Z/;
        my $Library = $1;
        next if(not $Library or (defined $TargetLibrary and $Library ne $TargetLibrary));
        print "[$Library]: ";
        if(create_target_descriptor($Library, $D_Path, 0)<0)
        {
            print "no\n";
            $NOT_INSTALLED .= "$Library\n";
            $Fail+=1;
        }
        else
        {
            print "yes\n";
            $Success+=1;
        }
    }
    print "installed: $Success/".($Success+$Fail)."\n";
    writeFile("results/not_installed", $NOT_INSTALLED);
    return $NOT_INSTALLED;
}

sub run_tests()
{#Running ABI Compliance Checker on two sets of library descriptors
    rmtree("compat_reports/", "target_descriptors/", "results/");
    mkpath("results/");
    my @Descriptors = cmd_find($Source."/descriptors","f","");
    my @ABIs = cmd_find($Source."/test_data","f","");
    my $JOURNAL = SimpleJournal->new("results/journal");
    $JOURNAL->test_name("ABI compliance test for $Distribution_Name and $TESTSUITE_DISTRIBUTION");
    $JOURNAL->tests_expected((defined $TargetLibrary)?1:$#ABIs+1);
    foreach my $D_Path (sort {lc($a) cmp lc($b)} @Descriptors)
    {
        $D_Path =~ /\/([^\/]*)\.desc\Z/;
        my $Library = $1;
        next if(not $Library or (defined $TargetLibrary and $Library ne $TargetLibrary));
        $JOURNAL->test_case_start($Library);
        my $ABI_Path = (cmd_find($Source."/test_data/$Library","f","$Library\_*abi\.tar\.gz"))[0];
        next if(not $ABI_Path);
        print "creating descriptor for $Library\n";
        my $Res = create_target_descriptor($Library, $D_Path, 1);
        if($Res>0)
        {
            print "testing for $Library ABI compliance\n";
            my $Int_List_Path = $Source."/int_list/".$Library.".ints";
            print "running abi-compliance-checker.pl -l \'$Library\' -d1 \'target_descriptors/$Library\.desc\' -d2 \'$ABI_Path\' -int_list \'$Int_List_Path\'\n";
            system("$FindBin::Bin/abi-compliance-checker.pl -l \'$Library\' -d1 \'target_descriptors/$Library\.desc\' -d2 \'$ABI_Path\' -int_list \'$Int_List_Path\'");
            print "\n";
            check_result($Library);
            my $Report_Path = (cmd_find("compat_reports/$Library","f","abi_compat_report.html"))[0];
            $JOURNAL->test_description("{noreplace}Test for library <a href=\"./$Report_Path\">$Library ABI compliance</a>{/noreplace}");
            my $Result = $TestResult{$Library}{"Withdrawn"}+$TestResult{$Library}{"T_High"}+$TestResult{$Library}{"T_Medium"}+$TestResult{$Library}{"I_High"}+$TestResult{$Library}{"I_Medium"};
            if($Result==0)
            {
                $JOURNAL->test_case_result("passed");
            }
            else
            {
                my $Comment = "{noreplace}".create_summary($TestResult{$Library}{"Withdrawn"}, $TestResult{$Library}{"T_High"}, $TestResult{$Library}{"T_Medium"}, $TestResult{$Library}{"T_Low"}, $TestResult{$Library}{"I_High"}, $TestResult{$Library}{"I_Medium"}, $TestResult{$Library}{"I_Low"}, $TestResult{$Library}{"Constants"})."{/noreplace}";
                if($Res==2)
                {
                    $Comment = "<span style='color:Red;'>Attention</span>: There are more than one versions of the shared library found in the system, so test results may be incorrect\n".$Comment;
                    $JOURNAL->test_comment($Comment);
                    $JOURNAL->test_case_result("warning");
                }
                else
                {
                    $JOURNAL->test_comment($Comment);
                    $JOURNAL->test_case_result("failed");
                }
            }
            $TestResult{$Library}{"IsCorrect"} = 1;
        }
        else
        {
            $JOURNAL->test_description("Test for library $Library ABI compliance");
            my $Comment = get_comment($Res, $Library);
            print "ERROR: $Comment\n";
            $JOURNAL->test_comment($Comment);
            $JOURNAL->test_case_result("untested");
            $TestResult{$Library}{"IsCorrect"} = 0;
            $TestResult{$Library}{"Comment"} = $Comment;
        }
        print "\n";
    }
    $JOURNAL->close();
}

sub get_comment($$)
{
    my ($ErrCode, $Library) = @_;
    if($ErrCode==-1)
    {
        return "necessary header files for $Library were not found";
    }
    elsif($ErrCode==-2)
    {
        return "necessary shared objects for $Library were not found";
    }
    elsif($ErrCode==-3)
    {
        return "necessary include paths for $Library were not found";
    }
    elsif($ErrCode==-4)
    {
        return "necessary shared object with soname ".$LibName_SoName{$Library}." was not found";
    }
    else
    {
        return "necessary header files or shared objects for $Library were not found";
    }
}

sub word_tr($$)
{
    my ($word, $count) = @_;
    if($count == 1)
    {
        return $word;
    }
    else
    {
        return $word."s";
    }
}

sub create_index()
{
    my $CssStyles = "<style type=\"text/css\">
    hr{color:Black;background-color:Black;height:1px;border: 0;}
    td.table_header{background-color:#eeeeee;}
    td.summary_item{font-size:15px;font-family:Arial;}
    td.summary_item_value{padding-right:5px;width:50px;}
    a.report_link{color:Blue;font-size:12px;}
    td.warnings{background-color:#F4F4AF;}
    td.failures{background-color:#FFC3CE;}
    td.success{background-color:#CCFFCC;}</style>\n";
    my $LIBS_INFO = "<h2 style='margin-bottom:0px;padding-bottom:0px;'>Details for Libraries (".keys(%TestResult).")</h2><hr/>\n<table cellpadding='3' border='1' style='border-collapse:collapse;'>";
    $LIBS_INFO .= "<tr><td rowspan='2' class='table_header'><b>Library</b></td><td rowspan='2' class='table_header' align='center'><b>Absent Interfaces<br/>in $TESTSUITE_DISTRIBUTION</b></td><td colspan='3' class='table_header' align='center'><b>Problems in Data Types</b></td><td colspan='3' class='table_header' align='center'><b>Interface Problems</b></td><td rowspan='2' class='table_header' align='center'><b>Constants<br/>Problems</b></td></tr>\n<tr><td class='table_header' style='color:Red;' align='center'>High risk</td><td class='table_header' align='center'>Medium risk</td><td class='table_header' align='center'>Warnings</td><td class='table_header' style='color:Red;' align='center'>High risk</td><td class='table_header' align='center'>Medium risk</td><td class='table_header' align='center'>Warnings</td></tr>\n";
    my ($Total_Libs, $Total_Miss) = (0, 0);
    my ($Withdrawn_Total, $T_High_Total, $T_Medium_Total, $T_Low_Total, $I_High_Total, $I_Medium_Total, $I_Low_Total, $Constants_Total) = (0, 0, 0, 0, 0, 0, 0, 0);
    foreach my $Library (sort {lc($a) cmp lc($b)} keys(%TestResult))
    {
        if(not $TestResult{$Library}{"IsCorrect"})
        {
            $Total_Miss += 1;
            $LIBS_INFO .= "<tr><td><a>$Library</a></td><td colspan='8' style='color:Red;'>&nbsp;&nbsp;&nbsp;".$TestResult{$Library}{"Comment"}."</td></tr>\n";
            next;
        }
        next if(not -e "compat_reports/$Library");
        $Total_Libs += 1;
        $Withdrawn_Total += $TestResult{$Library}{"Withdrawn"};
        $T_High_Total += $TestResult{$Library}{"T_High"};
        $T_Medium_Total += $TestResult{$Library}{"T_Medium"};
        $T_Low_Total += $TestResult{$Library}{"T_Low"};
        $I_High_Total += $TestResult{$Library}{"I_High"};
        $I_Medium_Total += $TestResult{$Library}{"I_Medium"};
        $I_Low_Total += $TestResult{$Library}{"I_Low"};
        $Constants_Total += $TestResult{$Library}{"Constants"};
        
        my $Report_Path = (cmd_find("compat_reports/$Library","f","abi_compat_report.html"))[0];
        
        my $Withdrawn_Link = "<td class='success' align='center'>0</td>";
        $Withdrawn_Link = "<td class='failures' align='center'><a class='report_link' href='$Report_Path#Withdrawn'>".$TestResult{$Library}{"Withdrawn"}." absent</a></td>" if($TestResult{$Library}{"Withdrawn"}>0);
        
        my $T_High_Count = $TestResult{$Library}{"T_High"};
        my $T_High_Link = "<td class='success' align='center'>0</td>";
        $T_High_Link = "<td class='failures' align='center'><a class='report_link' href='$Report_Path#Type_Problems_High'>$T_High_Count ".word_tr("problem", $T_High_Count)."</a></td>" if($T_High_Count>0);
        
        my $T_Medium_Count = $TestResult{$Library}{"T_Medium"};
        my $T_Medium_Link = "<td class='success' align='center'>0</td>";
        $T_Medium_Link = "<td class='failures' align='center'><a class='report_link' href='$Report_Path#Type_Problems_Medium'>$T_Medium_Count ".word_tr("problem", $T_Medium_Count)."</a></td>" if($T_Medium_Count>0);
        
        my $T_Low_Count = $TestResult{$Library}{"T_Low"};
        my $T_Low_Link = "<td class='success' align='center'>0</td>";
        $T_Low_Link = "<td class='warnings' align='center'><a class='report_link' href='$Report_Path#Type_Problems_Low'>$T_Low_Count ".word_tr("warning", $T_Low_Count)."</a></td>" if($T_Low_Count>0);
        
        my $I_High_Count = $TestResult{$Library}{"I_High"};
        my $I_High_Link = "<td class='success' align='center'>0</td>";
        $I_High_Link = "<td class='failures' align='center'><a class='report_link' href='$Report_Path#Interface_Problems_High'>$I_High_Count ".word_tr("problem", $I_High_Count)."</a></td>" if($I_High_Count>0);
        
        my $I_Medium_Count = $TestResult{$Library}{"I_Medium"};
        my $I_Medium_Link = "<td class='success' align='center'>0</td>";
        $I_Medium_Link = "<td class='failures' align='center'><a class='report_link' href='$Report_Path#Interface_Problems_Medium'>$I_Medium_Count ".word_tr("problem", $I_Medium_Count)."</a></td>" if($I_Medium_Count>0);
        
        my $I_Low_Count = $TestResult{$Library}{"I_Low"};
        my $I_Low_Link = "<td class='success' align='center'>0</td>";
        $I_Low_Link = "<td class='warnings' align='center'><a class='report_link' href='$Report_Path#Interface_Problems_Low'>$I_Low_Count ".word_tr("warning", $I_Low_Count)."</a></td>" if($I_Low_Count>0);
        
        my $Constants_Count = $TestResult{$Library}{"Constants"};
        my $Constants_Link = "<td class='success' align='center'>0</td>";
        $Constants_Link = "<td class='warnings' align='center'><a class='report_link' href='$Report_Path#Constants'>$Constants_Count ".word_tr("warning", $Constants_Count)."</a></td>" if($Constants_Count>0);
        
        $LIBS_INFO .= "<tr><td class='table_header'><a style='color:Blue;' href=\'$Report_Path\'>$Library</a></td>".$Withdrawn_Link.$T_High_Link.$T_Medium_Link.$T_Low_Link.$I_High_Link.$I_Medium_Link.$I_Low_Link.$Constants_Link."</tr>\n";
    }
    $LIBS_INFO .= "</table>\n";
    
    my $Total_Problems = $Withdrawn_Total+$T_High_Total+$T_Medium_Total+$I_High_Total+$I_Medium_Total;
    my $Total_Warnings = $T_Low_Total+$I_Low_Total+$Constants_Total;
    my $SUMMARY = "<h2 style='margin-bottom:0px;padding-bottom:0px;'>Summary</h2><hr/>\n<table cellpadding='3' border='1' style='border-collapse:collapse;'>\n";
    $SUMMARY .= "<tr><td class='table_header summary_item'>Total libraries checked</td><td align='right' class='summary_item_value'>$Total_Libs</td></tr>";
    $SUMMARY .= "<tr><td class='table_header summary_item'>Total problems found</td><td align='right' class='summary_item_value'>$Total_Problems</td></tr>";
    $SUMMARY .= "<tr><td class='table_header summary_item'>Total warnings</td><td align='right' class='summary_item_value'>$Total_Warnings</td></tr>";
    $SUMMARY .= "</table>\n";
    
    my $PROBLEM_SUMMARY = "<h2 style='margin-bottom:0px;padding-bottom:0px;'>Problem Summary</h2><hr/>\n";
    $PROBLEM_SUMMARY .= create_summary($Withdrawn_Total, $T_High_Total, $T_Medium_Total, $T_Low_Total, $I_High_Total, $I_Medium_Total, $I_Low_Total, $Constants_Total);
    
    my $INDEX = "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n<head>\n<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />
    <title>ABI compliance report for $Distribution_Name and $TESTSUITE_DISTRIBUTION</title>\n<!--Styles-->\n".$CssStyles."\n<!--Styles_End-->\n</head>\n<body><h1>ABI compliance report for the <span style='color:Blue;white-space:nowrap;'>$Distribution_Name</span> and <span style='color:Blue;white-space:nowrap;'>$TESTSUITE_DISTRIBUTION</span> libraries</h1>\n$SUMMARY\n$PROBLEM_SUMMARY\n$LIBS_INFO\n</body></html>";
    open(INDEX, ">results/abi_compliance_report.html");
    print INDEX $INDEX;
    close(INDEX);
    print "see the ABI compliance report in file 'results/abi_compliance_report.html'\n";
}

sub check_result($)
{
    my $Library = $_[0];
    return if(not -e "compat_reports/$Library");
    my $Report_Path = (cmd_find("compat_reports/$Library","f","abi_compat_report.html"))[0];
    return if(not $Report_Path);
    my $ABI_COMPAT_REPORT = readFile($Report_Path);
    my ($Withdrawn, $T_High, $T_Medium, $T_Low, $I_High, $I_Medium, $I_Low, $Constants) = (0, 0, 0, 0, 0, 0, 0, 0);
    if($ABI_COMPAT_REPORT =~ /\<input id='Withdrawn_Interfaces_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $Withdrawn = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Type_Problems_High_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $T_High = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Type_Problems_Medium_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $T_Medium = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Type_Problems_Low_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $T_Low = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Interface_Problems_High_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $I_High = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Interface_Problems_Medium_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $I_Medium = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Interface_Problems_Low_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $I_Low = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Changed_Constants_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $Constants = $1;
    }
    $TestResult{$Library} = {"Withdrawn"=>$Withdrawn, "T_High"=>$T_High, "T_Medium"=>$T_Medium, "T_Low"=>$T_Low, "I_High"=>$I_High, "I_Medium"=>$I_Medium, "I_Low"=>$I_Low, "Constants"=>$Constants};
}

sub create_summary($$$$$$$$)
{
    my ($Withdrawn, $T_High, $T_Medium, $T_Low, $I_High, $I_Medium, $I_Low, $Constants) = @_;
    my $SUMMARY = "<table cellpadding='3' border='1' style='border-collapse:collapse;'>";
    $SUMMARY .= "<tr><td class='table_header summary_item' colspan='2'>Absent Interfaces<br/>in $TESTSUITE_DISTRIBUTION</td><td align='right' class='summary_item_value'>$Withdrawn</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item' colspan='2'>Constants Problems</td><td align='right' class='summary_item_value'>$Constants</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item' rowspan='3'>Problems in<br/>Data Types</td><td class='table_header summary_item' style='color:Red;'>High risk</td><td align='right' class='summary_item_value'>$T_High</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item'>Medium risk</td><td align='right' class='summary_item_value'>$T_Medium</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item'>Warnings</td><td align='right' class='summary_item_value'>$T_Low</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item' rowspan='3'>Interface<br/>Problems</td><td class='table_header summary_item' style='color:Red;'>High risk</td><td align='right' class='summary_item_value'>$I_High</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item'>Medium risk</td><td align='right' class='summary_item_value'>$I_Medium</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item'>Warnings</td><td align='right' class='summary_item_value'>$I_Low</td></tr>";
    $SUMMARY .= "</table>";
    return $SUMMARY;
}

sub toUnix($)
{
    my $Text = $_[0];
    $Text=~s/\r//g;
    return $Text;
}

sub readFile($)
{
    my $Path = $_[0];
    return "" if(not $Path or not -f $Path);
    open (FILE, $Path);
    my $Content = join("", <FILE>);
    close(FILE);
    return toUnix($Content);
}

sub readDescriptor($$)
{
    my ($Library, $Descriptor_Path) = @_;
    return if(not $Descriptor_Path or not $Library);
    my $Descriptor_File = readFile($Descriptor_Path);
    $Descriptor_File =~ s|/\*[^*]*\*/||igs;
    if(not $Descriptor_File)
    {
        print "WARNING: descriptor $Descriptor_Path is empty\n";
        return;
    }
    if($Descriptor_File =~ m/<version>[ \n]*(.*?)[ \n]*<\/version>/ios)
    {
        $Descriptor{$Library}{"Version"} = $1;
    }
    else
    {
        print "WARNING: version is not specified in the $Descriptor_Path\n";
    }
    if($Descriptor_File =~ m/<headers>[ \n]*(.*?)[ \n]*<\/headers>/ios)
    {
        $Descriptor{$Library}{"Headers"} = $1;
    }
    else
    {
        print "WARNING: header files were not specified in the $Descriptor_Path\n";
    }
    if($Descriptor_File =~ m/<libs>[ \n]*(.*?)[ \n]*<\/libs>/ios)
    {
        $Descriptor{$Library}{"Libs"} = $1;
    }
    else
    {
        print "WARNING: shared objects were not specified in the $Descriptor_Path\n";
    }
    if($Descriptor_File =~ m/<include_paths>[ \n]*(.*?)[ \n]*<\/include_paths>/ios)
    {
        $Descriptor{$Library}{"Include_Paths"} = $1;
    }
    if($Descriptor_File =~ m/<gcc_options>[ \n]*(.*?)[ \n]*<\/gcc_options>/ios)
    {
        $Descriptor{$Library}{"Gcc_Options"} = $1;
    }
    if($Descriptor_File =~ m/<opaque_types>[ \n]*(.*?)[ \n]*<\/opaque_types>/ios)
    {
        $Descriptor{$Library}{"Opaque_Types"} = $1;
    }
    if($Descriptor_File =~ m/<skip_interfaces>[ \n]*(.*?)[ \n]*<\/skip_interfaces>/ios)
    {
        $Descriptor{$Library}{"Skip_Interfaces"} = $1;
    }
    if($Descriptor_File =~ m/<include_preamble>[ \n]*(.*?)[ \n]*<\/include_preamble>/ios)
    {
        $Descriptor{$Library}{"Include_Preamble"} = $1;
    }
}

sub create_target_descriptor($$$)
{
    mkpath("target_descriptors");
    my ($Library, $Descriptor_Path, $Warnings) = @_;
    my $ReturnCode = 1;
    readDescriptor($Library, $Descriptor_Path);
    my $TARGET_DESCRIPTOR = "<version>\n    $Distribution_Name\n</version>\n\n";
    my $HEADERS = "";
    foreach my $Header_Path (split("\n", $Descriptor{$Library}{"Headers"}))
    {
        $Header_Path=~s/\A[ ]+//g;
        $Header_Path=~s/[ ]+\Z//g;
        if(my $New_Path = findPath($Header_Path, "f"))
        {
            $HEADERS .= "    $New_Path\n";
        }
        elsif(my $New_Path = findPath($Header_Path, "d"))
        {
            $HEADERS .= "    $New_Path\n";
        }
        else
        {
            print "ERROR: corresponding header file for \'$Header_Path\' was not found\n" if($Warnings);
        }
    }
    return -1 if(not $HEADERS);
    $TARGET_DESCRIPTOR .= "<headers>\n$HEADERS</headers>\n\n";
    
    my $LIBS = "";
    foreach my $Lib_Path (split("\n", $Descriptor{$Library}{"Libs"}))
    {
        $Lib_Path=~s/\A[ ]+//g;
        $Lib_Path=~s/[ ]+\Z//g;
        if(my $New_Path = findPath($Lib_Path, "f"))
        {
            my @Family = findFamily(get_FileName($Lib_Path));
            if($#Family==0)
            {
                if(compare_sonames($Lib_Path, $New_Path)
                or -f "/usr/lib/".$LibPath_SoName{$Lib_Path}
                or -f "/lib/".$LibPath_SoName{$Lib_Path})
                {# FIXME: check other linker paths
                    $LIBS .= "    $New_Path\n";
                }
                else
                {
                    return -2;
                }
            }
            elsif($#Family>0)
            {
                my $Founded = "";
                foreach my $Path (sort {length($a)<=>length($b)} @Family)
                {
                    if(compare_sonames($Lib_Path, $Path)
                    or -f "/usr/lib/".$LibPath_SoName{$Lib_Path}
                    or -f "/lib/".$LibPath_SoName{$Lib_Path})
                    {# FIXME: check other linker paths
                        $Founded = $Path;
                    }
                }
                if($Founded)
                {
                    $ReturnCode = 2;
                    print "WARNING: There are more than one versions of the shared library found in the system";
                    $LIBS .= "    $Founded\n";
                }
                else
                {
                    return -4;
                }
            }
        }
        elsif(my $New_Path = findPath($Lib_Path, "d"))
        {
            $LIBS .= "    $New_Path\n";
        }
        else
        {
            print "ERROR: corresponding shared object for \'$Lib_Path\' was not found\n" if($Warnings);
        }
    }
    return -2 if(not $LIBS);
    $TARGET_DESCRIPTOR .= "<libs>\n$LIBS</libs>\n\n";
    
    my $INCLUDE_PATHS = "";
    foreach my $Include_Path (split("\n", $Descriptor{$Library}{"Include_Paths"}))
    {
        $Include_Path=~s/\A[ ]+//g;
        $Include_Path=~s/[ ]+\Z//g;
        if(my $New_Path = findPath($Include_Path, "d"))
        {
            if($New_Path ne $Include_Path)
            {
                if(my $AddPath = get_Directory($New_Path))
                {
                    if($AddPath ne "/usr/include")
                    {
                        $INCLUDE_PATHS .= "    $AddPath\n";
                    }
                }
            }
            $INCLUDE_PATHS .= "    $New_Path\n";
        }
        else
        {
            print "ERROR: corresponding include path for \'$Include_Path\' was not found\n" if($Warnings);
        }
    }
    return -3 if(not $INCLUDE_PATHS and $Descriptor{$Library}{"Include_Paths"});
    $TARGET_DESCRIPTOR .= "<include_paths>\n$INCLUDE_PATHS</include_paths>\n\n" if($INCLUDE_PATHS);
    
    $TARGET_DESCRIPTOR .= "<gcc_options>\n    ".$Descriptor{$Library}{"Gcc_Options"}."\n</gcc_options>\n\n" if($Descriptor{$Library}{"Gcc_Options"});
    $TARGET_DESCRIPTOR .= "<opaque_types>\n    ".$Descriptor{$Library}{"Opaque_Types"}."\n</opaque_types>\n\n" if($Descriptor{$Library}{"Opaque_Types"});
    $TARGET_DESCRIPTOR .= "<skip_interfaces>\n    ".$Descriptor{$Library}{"Skip_Interfaces"}."\n</skip_interfaces>\n\n" if($Descriptor{$Library}{"Skip_Interfaces"});
    $TARGET_DESCRIPTOR .= "<include_preamble>\n    ".$Descriptor{$Library}{"Include_Preamble"}."\n</include_preamble>\n\n" if($Descriptor{$Library}{"Include_Preamble"});
    writeFile("target_descriptors/$Library.desc", $TARGET_DESCRIPTOR);
    return $ReturnCode;
}

sub get_soname($)
{
    my $Path = $_[0];
    return "" if(not -f $Path);
    return $Cache{"get_soname"}{$Path} if($Cache{"get_soname"}{$Path});
    if(`readelf -Wa $Path|grep SONAME`=~/library soname:[ ]*\[(.+)\]/i)
    {
        $Cache{"get_soname"}{$Path}=$1;
        return $1;
    }
    else
    {
        return "";
    }
}

sub compare_sonames($$)
{
    my ($Path1, $Path2) = @_;
    return 0 if(not -f $Path2);
    my $SoName1 = $LibPath_SoName{$Path1};
    if(my $SoName2 = get_soname($Path2))
    {
        return ($SoName1 eq $SoName2);
    }
    else
    {
        return 0;
    }
}

sub get_FileName($)
{
    my $Destination = $_[0];
    if($Destination =~ /\A(.*\/)([^\/]*)\Z/)
    {
        return $2;
    }
    else
    {
        return $Destination;
    }
}

sub get_Directory($)
{
    my $Destination = $_[0];
    return "" if($Destination=~m*\A\./*);
    if($Destination=~/\A(.*)\/([^\/]*)\Z/)
    {
        return $1;
    }
    else
    {
        return "";
    }
}

sub writeFile($$)
{
    my ($Path, $Content) = @_;
    return if(not $Path);
    open (FILE, ">".$Path) || die ("Can't open file \'$Path\': $!\n");
    print FILE $Content;
    close(FILE);
}

sub findPath($$)
{
    my ($Path, $Type) = @_;
    return "" if(not $Path);
    if($Path!~/\// or (($Type eq "d")?-d $Path:-f $Path))
    {
        return $Path;
    }
    else
    {
        findDevelFiles_1stStage();
        if(my $NewPath = searchPath($Path, $Type))
        {
            return $NewPath;
        }
    }
    return "";
}

sub findDevelFiles_1stStage()
{
    return if(keys(%DevelFiles));
    foreach my $Path (sort {lc($a) cmp lc($b)} split("\n", `find /usr/include /usr/lib /lib`))
    {
        chomp($Path);
        if(-d $Path)
        {
            $DevelFiles{$Path} = "d";
        }
        else
        {
            $DevelFiles{$Path} = "f";
        }
        my $FileName = get_FileName($Path);
        $DevelFileName_Paths{$FileName}{$Path} = 1;
        $DevelFileName_Paths{remove_suffix($FileName)}{$Path} = 1;
    }
}

sub remove_suffix($)
{
    my $Path = $_[0];
    $Path=~s![0-9.-_]*\Z!!g;
    return $Path;
}

sub searchPath($$)
{
    my ($Path, $Type) = @_;
    return "" if(not $Path);
    if(my $Res = searchEntity($Path, $Type))
    {
        return $Res;
    }
    elsif($Type eq "f")
    {
        if($Path =~ s!\.so\.[0-9.]*\Z!.so!g)
        {#solibs
            if(my $Res = searchEntity($Path, "f"))
            {
                return $Res;
            }
        }
    }
    elsif($Type eq "d")
    {
        if($Path =~ s![0-9.-_]*\Z!!g)
        {#directories
            if(my $Res = searchEntity($Path, "d"))
            {
                return $Res;
            }
        }
    }
    else
    {
        return "";
    }
}

sub searchEntity($$)
{
    my ($Path, $Type) = @_;
    return "" if(not $Path);
    if(my $Res = search_in_devel_files($Path, $Type))
    {
        return $Res;
    }
    else
    {
        while($Path=~s!\A(.+?)/!!o and $Path)
        {
            if(my $Res = search_in_devel_files($Path, $Type))
            {
                return $Res;
            }
        }
    }
}

sub count_slash($)
{
    my $Code=$_[0];
    my $Count=0;
    while($Code=~s/\///)
    {
        $Count+=1;
    }
    return $Count;
}

sub search_in_devel_files($$)
{
    my ($Path, $Type) = @_;
    return "" if(not $Path);
    foreach my $DevelPath (sort {length($a)<=>length($b)} keys(%{$DevelFileName_Paths{get_FileName($Path)}}), sort {length($a)<=>length($b)} keys(%{$DevelFileName_Paths{remove_suffix(get_FileName($Path))}}))
    {
        next if($Type ne $DevelFiles{$DevelPath});
        if(substr($Path, 0, 1) eq "/")
        {
            if($DevelPath=~/\A$Path[0-9.-_]*\Z/)
            {
                return $DevelPath;
            }
        }
        else
        {
            if($DevelPath=~/\/$Path[0-9.-_]*\Z/)
            {
                return $DevelPath;
            }
        }
    }
}

sub findFamily($)
{#search for shared objects set with the same soname prefix
    findDevelFiles_1stStage();
    my $Path = $_[0];
    return () if(not $Path);
    my (%Family, %SoNames) = ();
    foreach my $DevelPath (sort {length($a)<=>length($b)} keys(%{$DevelFileName_Paths{get_FileName($Path)}}), sort {length($a)<=>length($b)} keys(%{$DevelFileName_Paths{remove_suffix(get_FileName($Path))}}))
    {
        if($DevelPath=~/\/$Path[0-9.]*/)
        {
            next if($SoNames{get_soname($DevelPath)});
            $Family{$DevelPath} = 1;
            $SoNames{get_soname($DevelPath)} = 1;
        }
    }
    my @FamilyArray = keys(%Family);
    return @FamilyArray;
}

sub detect_OS()
{
    my $Distribution = `lsb_release -s -d`;
    chomp($Distribution);
    $Distribution =~ s/^\s*\"(.*)\"\s*$/$1/;
    return $Distribution;
}

sub readSoNames($)
{
    my $FilePath = $_[0];
    foreach my $Line (split("\n", readFile($FilePath)))
    {
        if($Line=~/\A([^;]+);([^;]+)\Z/)
        {
            my ($Path, $SoName) = ($1, $2);
            $LibPath_SoName{$Path} = $SoName;
            $Path=~s/.so.*\Z//;
            $LibName_SoName{get_FileName($Path)} = $SoName;
        }
    }
}

sub scenario()
{
    if(defined $Help)
    {
        HELP_MESSAGE();
        exit(0);
    }
    if(not $Distribution_Name)
    {
        print "selecting distribution name ... ";
        $Distribution_Name = detect_OS();
        if($Distribution_Name)
        {
            print $Distribution_Name."\n";
        }
        else
        {
            print "can't select\nselect distribution name (d <name> option)\n";
            exit(1);
        }
    }
    if(defined $Check_Devel_Files)
    {
        if(not $Source)
        {
            print "select test suite source path (-src <path> option)\n";
            exit(1);
        }
        if(-f $Source."/sonames")
        {
            readSoNames($Source."/sonames");
        }
        else
        {
            print "ERROR: file 'sonames' was not found in the directory \'$Source\'\n";
            exit(1);
        }
        if(not -e $Source."/descriptors/")
        {
            print "ERROR: library descriptors were not found in the directory '$Source/descriptors/'\n";
            exit(1);
        }
        exit(check_devel_files() ne "");
    }
    if(defined $Create_TestSuite)
    {
        if(not $Source)
        {
            print "select test suite source path (-src <path> option)\n";
            exit(1);
        }
        if(not -f $FindBin::Bin."/abi-compliance-checker.pl")
        {
            print "'abi-compliance-checker.pl' was not found\n";
            exit(1);
        }
        if(not -d $Source."/int_list/")
        {
            print "ERROR: interface lists were not found in the directory '$Source/int_list/'\n";
            exit(1);
        }
        if(not -d $Source."/descriptors/")
        {
            print "ERROR: library descriptors were not found in the directory '$Source/descriptors/'\n";
            exit(1);
        }
        if($TargetLibrary)
        {
            recreate_test_case($TargetLibrary);
        }
        else
        {
            create_test_suite();
        }
    }
    if(defined $Run_Tests)
    {
        if(not $Source)
        {
            print "select test suite source path (-src <path> option)\n";
            exit(1);
        }
        if(-f $Source."/sonames")
        {
            readSoNames($Source."/sonames");
        }
        else
        {
            print "ERROR: file 'sonames' was not found in the directory \'$Source\'\n";
            exit(1);
        }
        if(-f $Source."/distribution.name")
        {
            $TESTSUITE_DISTRIBUTION = readFile($Source."/distribution.name");
            chomp($TESTSUITE_DISTRIBUTION);
            if(not $TESTSUITE_DISTRIBUTION)
            {
                print "ERROR: distribution name is not set in the file '$Source/distribution.name'\n";
                exit(1);
            }
        }
        else
        {
            print "file '$Source/distribution.name' is not exists\n";
            exit(1);
        }
        if(not -f $FindBin::Bin."/abi-compliance-checker.pl")
        {
            print "'abi-compliance-checker.pl' was not found\n";
            exit(1);
        }
        if(not -d $Source."/int_list/")
        {
            print "ERROR: interfaces lists were not found in the directory '$Source/int_list/'\n";
            exit(1);
        }
        if(not -d $Source."/descriptors/")
        {
            print "ERROR: library descriptors were not found in the directory '$Source/descriptors/'\n";
            exit(1);
        }
        if(not -d $Source."/test_data/")
        {
            print "ERROR: test data was not found in the directory '$Source/test_data/'\n";
            exit(1);
        }
        rmtree("results/");
        mkpath("results/");
        run_tests();
        create_index();
        system("mv", "compat_reports", "results/compat_reports") if(-d "compat_reports");
    }
}

scenario();
